/* X-Trans test file creator, in FITS format because it's easier to write than RAW
 * Based on code from Siril - https://siril.org
 * License: BSD
 * Author: Vincent Hourdin
 */
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fitsio.h>
// build with gcc -Wall -o xtrans_creator `pkg-config --cflags cfitsio` xtrans_creator.c `pkg-config --libs cfitsio`

const char pattern[] =
"GGRGGB"
"GGBGGR"
"BRGRBG"
"GGBGGR"
"GGRGGB"
"RBGBRG";

#define IMAGE_SIZE 100L
#define RED_VALUE 50000
#define GREEN_VALUE 0
#define BLUE_VALUE 0
#define BOTTOM_UP 1	// 0 for top-down

void report_fits_error(int status) {
	if (status) {
		char errmsg[FLEN_ERRMSG];
		while (fits_read_errmsg(errmsg)) {
			printf("FITS error: %s\n", errmsg[0] != '\0' ? errmsg : "unknown" );
		}
	}
}

/* convert the string-described X-Trans pattern into an int array with value corresponding to filter */
static int compile_XTrans_pattern(const char *bayer, unsigned int xtrans[6][6]) {
        if (strlen(bayer) != 36) {
                printf("FITS header does not contain a proper XTRANS pattern, demosaicing cannot be done\n");
                return 1;
        }
        int i = 0;
        for (int y = 0; y < 6; y++) {
                for (int x = 0; x < 6; x++) {
                        switch (bayer[i]) {
				case 'R':
					xtrans[y][x] = 0;
					break;
				case 'G':
					xtrans[y][x] = 1;
					break;
				case 'B':
					xtrans[y][x] = 2;
					break;
				default:
					return 1;
			}
			i++;
		}
	}
	return 0;
}

int main() {
	fitsfile *fptr;
	int status = 0;
	unlink("test.fits");
	fits_create_diskfile(&fptr, "test.fits", &status);
	if (status) {
		report_fits_error(status);
		return 1;
	}
	long naxes[] = { IMAGE_SIZE, IMAGE_SIZE };
	if (fits_create_img(fptr, USHORT_IMG, 2, naxes, &status)) {
		report_fits_error(status);
		return 1;
	}

	double zero = 32768.0;
	double scale = 1.0;
#if BOTTOM_UP
	char row_order[] = "BOTTOM-UP";
#else
	char row_order[] = "TOP-DOWN";
#endif
	char program[] = "Siril X-Trans generator";

	status = 0;
        fits_update_key(fptr, TDOUBLE, "BZERO", &zero,
                        "offset data range to that of unsigned short", &status);

	fits_update_key(fptr, TDOUBLE, "BSCALE", &scale, "default scaling factor",
                        &status);

	fits_update_key(fptr, TSTRING, "ROWORDER", &row_order,
                                "Order of the rows in image array", &status);

	fits_update_key(fptr, TSTRING, "BAYERPAT", (char*)pattern,
			"X-Trans filter array pattern", &status);

	fits_update_key(fptr, TSTRING, "PROGRAM", program,
                        "Software that created this HDU", &status);

	if (status) {
		report_fits_error(status);
		return 1;
	}

	unsigned short *buf = malloc(IMAGE_SIZE * IMAGE_SIZE * sizeof(unsigned short));
	if (!buf) return 1;

	unsigned int xtrans[6][6];
	compile_XTrans_pattern(pattern, xtrans);

	// fill the image top-down
	int i = 0;
	for (int y = 0; y < IMAGE_SIZE; y++) {
		int pattern_y = y % 6;
		for (int x = 0; x < IMAGE_SIZE; x++) {
			int pattern_x = x % 6;
			unsigned short pixval;
			switch (xtrans[pattern_y][pattern_x]) {
				case 0:
					pixval = RED_VALUE;
					break;
				case 1:
					pixval = GREEN_VALUE;
					break;
				case 2:
					pixval = BLUE_VALUE;
					break;
			}
			buf[i++] = pixval;
		}
	}

#if BOTTOM_UP
	for (int y = 0; y < IMAGE_SIZE/2; y++) {
		unsigned short tmp[IMAGE_SIZE];
		memcpy(tmp, buf+y*IMAGE_SIZE, IMAGE_SIZE*sizeof(unsigned short));
		memcpy(buf+y*IMAGE_SIZE, buf+(IMAGE_SIZE-y-1)*IMAGE_SIZE, IMAGE_SIZE*sizeof(unsigned short));
		memcpy(buf+(IMAGE_SIZE-y-1)*IMAGE_SIZE, tmp, IMAGE_SIZE*sizeof(unsigned short));
	}
#endif

	long orig[3] = { 1L, 1L, 1L };
	if (fits_write_pix(fptr, TUSHORT, orig, IMAGE_SIZE * IMAGE_SIZE, buf, &status)) {
		report_fits_error(status);
		return 1;
	}

	if (fits_close_file(fptr, &status)) {
		report_fits_error(status);
		return 1;
	}
	fptr = NULL;

	return 0;
}
